local LTC = LibStub("LibTrinityCore-1.0")
assert(LTC)

local datastore = {}
local function BuildDataStore(mapid, data)
  assert(#data > 0)
  local id, mapname = data[1]:match("^Listing all spawn points in map (%d+) %((.+)%)")
  assert(tonumber(id) == mapid)
  local store = {mapid = mapid, mapname = mapname}
  for i=2,#data do
    local spawnType, spawnId, entry, name, x, y, z = data[i]:match("^Type: ([01]) | SpawnId: (%d+) | Entry: (%d+) %((.*)%) | X: ([0-9%-%.]+) | Y: ([0-9%-%.]+) | Z: ([0-9%-%.]+)$")
    assert(spawnType)
    table.insert(store, {
      spawnType = tonumber(spawnType),
      spawnId = tonumber(spawnId),
      entry = tonumber(entry),
      name = name,
      posX = tonumber(x),
      posY = tonumber(y),
      posZ = tonumber(z)
    })
  end
  datastore[mapid] = store
end
local function GetDataStore(callback, force)
  LTC:DoCommand(("gps %s"):format((UnitName("player"))), function (ok, data)
    assert(ok)
    local mapid
    for _,l in ipairs(data) do
      mapid = tonumber((l:match("^Map: (%d+)")))
      if mapid then break end
    end
    assert(mapid)
    if (not force) and datastore[mapid] then
      callback(mapid, datastore[mapid])
    else
      LTC:DoCommand("list spawnpoints", function(ok, data) assert(ok) BuildDataStore(mapid, data) callback(mapid, datastore[mapid]) end)
    end
  end)
end

local redrawFrame

local function background(f,r,g,b,a, layer, sublayer)
  local tex = f:CreateTexture(nil,layer or "BACKGROUND", sublayer or 0)
  tex:SetTexture(r,g,b)
  tex:SetAlpha(a or 1.0)
  tex:SetAllPoints()
  return tex
end

local frame = CreateFrame("Frame", "SpawnGroupMapper", UIParent)
frame:SetWidth(500)
frame:SetHeight(400)
frame:SetMovable(true)
frame:SetResizable(true)
frame:SetMinResize(200, 150)
frame:SetClampedToScreen(true)
frame:SetPoint("TOP",UIParent,"CENTER")
frame:SetFrameStrata("HIGH")
background(frame,0.4,0.4,0.4,0.7)
frame:Hide()

local toggleMinimize
local header = CreateFrame("Frame", nil, frame)
header:SetHeight(30)
header:EnableMouse(true)
header:SetScript("OnMouseUp", function(self)
  self:GetParent():StopMovingOrSizing()
  local posx,posy = GetCursorPosition()
  if (abs(posx-self.posx)+abs(posy-self.posy)) < 1 then
    toggleMinimize()
  end
end)
header:SetScript("OnMouseDown", function(self) self.posx, self.posy = GetCursorPosition() self:GetParent():StartMoving() end)
background(header,0.4,0.4,1.0,0.7)
header:SetPoint("TOPLEFT", 5, -5)
header:SetPoint("TOPRIGHT", -5, -5)
header.text = header:CreateFontString(nil, "ARTWORK")
header.text:SetAllPoints()
header.text:SetFont("Fonts\\MORPHEUS.TTF", 13, "OUTLINE")
header.text:SetTextColor(0.8,0.8,0.8)
header.text:SetText("SpawnGroupMapper")

local scaler = CreateFrame("Frame", nil, frame)
scaler:SetPoint("BOTTOMRIGHT", -2, 2)
scaler:SetSize(10,10)
scaler:EnableMouse(true)
scaler.tex = background(scaler, 1, 1, 1, 0.3)
scaler.tex:Hide()
scaler:SetScript("OnMouseUp", function(self) self:GetParent():StopMovingOrSizing() redrawFrame() end)
scaler:SetScript("OnMouseDown", function(self) self:GetParent():StartSizing() end)
scaler:SetScript("OnEnter", function(self) self.tex:Show() end)
scaler:SetScript("OnLeave", function(self) self.tex:Hide() end)

local borders = {}
for i=1,4 do borders[i] = CreateFrame("Frame", nil, frame) background(borders[i], 0.3,0.3,0.3,1.0) end
for i=1,4,3 do borders[i]:SetPoint("TOPLEFT", -2, 2) end
for i=1,2,1 do borders[i]:SetPoint("TOPRIGHT", 2, 2) end
for i=2,3,1 do borders[i]:SetPoint("BOTTOMRIGHT", 2, -2) end
for i=3,4,1 do borders[i]:SetPoint("BOTTOMLEFT", -2, -2) end
for i=1,3,2 do borders[i]:SetHeight(4) end
for i=2,4,2 do borders[i]:SetWidth(4) end

local add = CreateFrame("Frame", nil, frame)
add:SetSize(24,24)
add:SetFrameLevel(header:GetFrameLevel()+1)
add:SetPoint("TOPRIGHT", header, "TOPRIGHT", -3, -3)
add:EnableMouse(true)
background(add, 0.4, 0.8, 0.4, 0.4)
add:Hide()
add:SetScript("OnMouseUp", function()
  local i = #frame.savedData.groups + 1
  frame.savedData.groups[i] = {
    id = i,
    name = ("Unnamed (#%d)"):format(i),
    collapsed = true
  }
  redrawFrame()
end)
add.text = add:CreateFontString(nil, "ARTWORK")
add.text:SetFont("FONTS\\skurri.TTF", 23, "THICKOUTLINE")
add.text:SetTextColor(0.2,0.7,0.2)
add.text:SetText("+")
add.text:SetPoint("CENTER")

local sql = CreateFrame("Button", nil, frame)
sql:SetSize(80, 24)
sql:SetFrameLevel(header:GetFrameLevel()+1)
sql:SetPoint("TOPLEFT", header, "TOPLEFT", 3, -3)
sql:RegisterForClicks("LeftButtonUp")
background(sql, 0.8, 0.4, 0.8, 0.7)
local generateSQL -- defined near end of file
sql:SetScript("OnClick", function() generateSQL() end)
sql:Hide()
sql.text = sql:CreateFontString(nil, "ARTWORK")
sql.text:SetFont("FONTS\\ARIALN.TTF", 15, "OUTLINE")
sql.text:SetTextColor(0.7,0.2,0.7)
sql.text:SetText("Export SQL")
sql.text:SetPoint("CENTER")

local scrollParent = CreateFrame("ScrollFrame", nil, frame)
scrollParent:SetPoint("TOPLEFT", 10, -42)
scrollParent:SetPoint("BOTTOMRIGHT", -10, 10)
local scrollChild = CreateFrame("Frame", nil, scrollParent)
--background(scrollChild,0.6,0.6,0.6,0.3)
scrollParent:SetScrollChild(scrollChild)
scrollParent:EnableMouseWheel(true)
scrollParent:SetScript("OnMouseWheel", function(_,delta)
  local diff = IsShiftKeyDown() and 75 or 10
  if delta < 0 then
    scrollParent:SetVerticalScroll(math.min(scrollParent:GetVerticalScrollRange(), scrollParent:GetVerticalScroll()+diff))
  elseif delta > 0 then
    scrollParent:SetVerticalScroll(math.max(0, scrollParent:GetVerticalScroll()-diff))
  end
end)

local dropdown = CreateFrame("Frame", nil, UIParent)
dropdown:SetFrameStrata("DIALOG")
dropdown.background = background(dropdown,0,0,0,1, "OVERLAY", 5)
dropdown.background:SetDrawLayer("OVERLAY",5)
dropdown:SetWidth(170)
local function dropdownentry_OnClick(e)
  local groupId = e.groupId
  local entry = dropdown.parent
  local currentGroup
  if entry.groupHeader then
    currentGroup = entry.groupHeader.data and entry.groupHeader.data.id or 0
  else
    currentGroup = entry.data and entry.data.id or 0
  end
  if entry.state == "SPAWNID" then
    frame.savedData.lookup[entry.data.spawnType][entry.data.spawnId] = groupId ~= 0 and groupId
  elseif entry.state == "ENTRYHEADER" then
    for _,c in ipairs(entry.data) do
      if currentGroup ~= 0 and frame.savedData.lookup[c.spawnType][c.spawnId] == currentGroup then
        frame.savedData.lookup[c.spawnType][c.spawnId] = groupId ~= 0 and groupId
      elseif currentGroup == 0 and not frame.savedData.lookup[c.spawnType][c.spawnId] then
        frame.savedData.lookup[c.spawnType][c.spawnId] = groupId ~= 0 and groupId
      end
    end
  elseif entry.state == "GROUPHEADER" then
    assert(currentGroup ~= 0)
    local mapping = {[-1] = -1}
    for i=1,#frame.savedData.groups do
      mapping[i] = i > currentGroup and i-1 or i
    end
    mapping[currentGroup] = groupId ~= 0 and groupId
    table.remove(frame.savedData.groups,currentGroup)
    for k,v in pairs(frame.savedData.lookup[0]) do
      frame.savedData.lookup[k] = mapping[v]
    end
    for k,v in pairs(frame.savedData.lookup[1]) do
      frame.savedData.lookup[k] = mapping[v]
    end
    for i=0,1 do
      for _,d in pairs(frame.data[i]) do
        local old = d.collapsed
        d.collapsed = setmetatable({},getmetatable(old))
        for k,v in pairs(old) do
          if k > currentGroup then
            d.collapsed[k-1] = v
          elseif k < currentGroup then
            d.collapsed[k] = v
          end
        end
      end
    end
  end
  redrawFrame() -- this also hides dropdown
end
dropdown.entries = setmetatable({},{__index=function(t,v)
  local e = CreateFrame("Frame", nil, dropdown)
  e:SetFrameStrata("DIALOG")
  if v == 1 then
    e:SetPoint("TOPLEFT", 5, -5)
    e:SetPoint("TOPRIGHT", -5, -5)
  else
    e:SetPoint("TOPLEFT", t[v-1], "BOTTOMLEFT", 0, -5)
    e:SetPoint("TOPRIGHT", t[v-1], "BOTTOMRIGHT", 0, -5)
  end
  e:SetHeight(21)
  e.text = e:CreateFontString(nil,"OVERLAY",7)
  e.text:SetPoint("LEFT")
  e.text:SetFont("FONTS\\ARIALN.TTF", 11, "OUTLINE")
  e.text:SetTextColor(0.8,0.8,0.8)
  background(e,0,0,0,0.2,"OVERLAY",6)
  e:EnableMouse(true)
  e:SetScript("OnMouseUp",dropdownentry_OnClick)
  t[v] = e
  return e
end})
function dropdown.Redraw()
  local entry = dropdown.parent
  local header = entry.groupHeader or entry
  local groupId = header.data and header.data.id or 0
  dropdown.background:SetTexture(entry.r,entry.g,entry.b)
  local j = 1
  local h = 5
  for i=-1,#frame.savedData.groups do
    if i ~= groupId then
      local n = i == 0 and "Unassigned" or i == -1 and "Ignored" or frame.savedData.groups[i].name
      local e = dropdown.entries[j]
      e.text:SetText(("#%d: %s"):format(i,n))
      e.groupId = i
      e:Show()
      j = j+1
      h = h+e:GetHeight()+5
    end
  end
  for i=j,#dropdown.entries do dropdown.entries[i]:Hide() end
  dropdown:SetHeight(h)
  dropdown:SetPoint("TOPLEFT", entry, "TOPRIGHT", 5, 0)
end

toggleMinimize = function()
  frame.minimized = not frame.minimized
  local offL = frame:GetLeft() - UIParent:GetLeft()
  local offT = frame:GetTop() - UIParent:GetTop()
  if frame.minimized then
    frame.fullHeight = frame:GetHeight()
    frame:SetHeight(40)
    header.text:SetText("SpawnGroupMapper (click to expand)")
    scaler:Hide()
    scrollParent:Hide()
    add:Hide()
    sql:Hide()
    dropdown:Hide()
    dropdown.parent = nil
  else
    frame:SetHeight(frame.fullHeight)
    header.text:SetText("SpawnGroupMapper")
    scaler:Show()
    scrollParent:Show()
    add:Show()
    sql:Show()
  end
  frame:ClearAllPoints()
  frame:SetPoint("TOPLEFT", UIParent, "TOPLEFT", offL, offT)
end

local function entry_SetState(entry, state)
  entry.state = state
  if state == "GROUPHEADER" then
    entry:SetHeight(29)
    entry.text:SetFont("FONTS\\ARIALN.TTF", 16, "THICKOUTLINE")
    entry.text:SetTextColor(0.45,0.45,0.45)
    entry.dropdown.text:SetTextColor(0.45,0.45,0.45)
    entry.text:SetPoint("LEFT", 2, 0)
  elseif state == "ENTRYHEADER" then
    entry:SetHeight(25)
    entry.text:SetFont("FONTS\\ARIALN.TTF", 13, "THICKOUTLINE")
    entry.text:SetTextColor(0.65,0.65,0.65)
    entry.dropdown.text:SetTextColor(0.65,0.65,0.65)
    entry.text:SetPoint("LEFT", 12, 0)
  elseif state == "SPAWNID" then
    entry:SetHeight(21)
    entry.text:SetFont("FONTS\\ARIALN.TTF", 11, "OUTLINE")
    entry.text:SetTextColor(0.8,0.8,0.8)
    entry.dropdown.text:SetTextColor(0.8,0.8,0.8)
    entry.text:SetPoint("LEFT", 22, 0)
  end
end
local function entry_openDropdown(entryDropdown)
  if dropdown.parent == entryDropdown.parent then
    dropdown:Hide()
    dropdown.parent = nil
  else
    dropdown:ClearAllPoints()
    dropdown.parent = entryDropdown.parent
    dropdown:Redraw()
    dropdown:Show()
  end
end
StaticPopupDialogs.SPAWNGROUPMAPPER_RENAME = {
  text = "Enter new name for group #%d: %s below:",
  button1 = ACCEPT,
  button2 = CANCEL,
  OnShow = function(self) self.button1:Disable() end,
  OnAccept = function(self)
    local new = self.editBox:GetText()
    if new:match("^%s*$") then return true end
    frame.savedData.groups[self.data].name = self.editBox:GetText()
    redrawFrame()
  end,
  whileDead = true,
  hideOnEscape = true,
  timeout = 30,
  EditBoxOnEnterPressed = function(self) self:GetParent().button1:Click() end,
  EditBoxOnTextChanged = function(self)
    local b = self:GetParent().button1
    b[self:GetText():len() > 0 and "Enable" or "Disable"](b)
  end,
  hasEditBox = true
}
local function entry_OnClick(entry,btn)
  if btn == "RightButton" then
    if entry.state == "GROUPHEADER" then
      if entry.data and entry.data.id > 0 then
        StaticPopup_Show("SPAWNGROUPMAPPER_RENAME", entry.data.id, entry.data.name).data = entry.data.id
      end
    end
  else
    if entry.state == "GROUPHEADER" then
      if entry.data then
        entry.data.collapsed = not entry.data.collapsed
        redrawFrame()
      end
    elseif entry.state == "ENTRYHEADER" then
      local group = entry.groupHeader.data and entry.groupHeader.data.id or 0
      entry.data.collapsed[group] = not entry.data.collapsed[group]
      redrawFrame()
    elseif entry.state == "SPAWNID" then
      LTC:DoCommand(("go creature %d"):format(entry.data.spawnId))
    end
  end
end
local children = setmetatable({},{__index=function(t,v)
  local entry = CreateFrame("Frame", nil, scrollChild)
  if v == 1 then
    entry:SetPoint("TOPLEFT", scrollChild, 5, -5)
    entry:SetPoint("TOPRIGHT", scrollChild, -5, 5)
  else
    entry:SetPoint("TOPLEFT", t[v-1], "BOTTOMLEFT", 0, -5)
    entry:SetPoint("TOPRIGHT", t[v-1], "BOTTOMRIGHT", 0, -5)
    t[v-1].next = entry
    entry.prev = t[v-1]
  end
  local r,g,b = math.random(), math.random(), math.random()
  entry.r,entry.g,entry.b = r,g,b
  entry.text = entry:CreateFontString(nil, "ARTWORK")
  entry.text:SetJustifyH("LEFT")
  entry.text:SetJustifyV("MIDDLE")
  entry.text:SetPoint("RIGHT", -175, 0)
  entry.text:SetWordWrap(false)
  entry.SetState = entry_SetState
  entry:EnableMouse(true)
  entry:SetScript("OnMouseUp", entry_OnClick)
  entry.dropdown = CreateFrame("Frame", nil, entry)
  entry.dropdown.parent = entry
  entry.dropdown:SetPoint("TOPRIGHT", -4, -4)
  entry.dropdown:SetPoint("BOTTOMRIGHT", -4, 4)
  entry.dropdown:SetWidth(170)
  entry.dropdown.text = entry.dropdown:CreateFontString(nil, "ARTWORK")
  entry.dropdown.text:SetFont("FONTS\\ARIALN.TTF", 11, "OUTLINE")
  entry.dropdown.text:SetPoint("LEFT")
  entry.dropdown.text:SetPoint("RIGHT", -15, 0)
  entry.dropdown.text:SetJustifyH("LEFT")
  entry.dropdown.text:SetJustifyV("MIDDLE")
  entry.dropdown.text:SetWordWrap(false)
  entry.dropdown.button = CreateFrame("Frame", nil, entry.dropdown)
  entry.dropdown.button:SetSize(13,13)
  entry.dropdown.button:SetPoint("RIGHT")
  entry.dropdown.button.text = entry.dropdown.button:CreateFontString(nil,"ARTWORK")
  entry.dropdown.button.text:SetFont("FONTS\\ARIALN.TTF", 9, "OUTLINE")
  entry.dropdown.button.text:SetTextColor(0.8,0.8,0.8)
  entry.dropdown.button.text:SetText("v")
  entry.dropdown.button.text:SetPoint("CENTER")
  entry.dropdown:EnableMouse(true)
  entry.dropdown:SetScript("OnMouseUp",entry_openDropdown)
  t[v] = entry
  return entry
end})

frame.showCreatures = true
frame.showGameObjects = false
frame.showIgnored = true
local ignoredDummy = {name="Ignored",id=-1,collapsed=true}
redrawFrame = function(mapid, data)
  mapid = mapid or frame.mapid
  assert(mapid)
  if data and data[1] then
    frame.mapid = data.mapid
    frame.mapname = data.mapname
    frame.data = {}
    for _,k in ipairs(data) do
      frame.data[k.spawnType] = frame.data[k.spawnType] or {}
      frame.data[k.spawnType][k.entry] = frame.data[k.spawnType][k.entry] or {collapsed = setmetatable({},{__index=function() return true end})}
      table.insert(frame.data[k.spawnType][k.entry], k)
    end
    add:Show()
    sql:Show()
  end
  local datas = {}
  if frame.showCreatures and frame.data[0] then table.insert(datas,frame.data[0]) end
  if frame.showGameObjects and frame.data[1] then table.insert(datas,frame.data[1]) end
  
  SpawnGroupMapper_data = SpawnGroupMapper_data or {}
  SpawnGroupMapper_data[mapid] = SpawnGroupMapper_data[mapid] or {groups={},lookup={[0]={},[1]={}}}
  frame.savedData = SpawnGroupMapper_data[mapid]
  local savedData = frame.savedData
  dropdown:Hide()
  dropdown.parent = nil
  local i = 1
  local h = 5
  for g=(frame.showIgnored and -1 or 0),#savedData.groups do
    local top = children[i]
    top:SetState("GROUPHEADER")
    if g <= 0 then
      top.text:SetText(g == 0 and "Unassigned" or "Ignored")
    else
      top.text:SetText(("Group #%d - %s"):format(g, savedData.groups[g].name))
    end
    top.dropdown.text:SetText(("#%d:%s"):format(g, g == 0 and "Unassigned" or g == -1 and "Ignored" or savedData.groups[g].name))
    top.dropdown[g == 0 and "Hide" or "Show"](top.dropdown)
    if g > 0 then
      top.data = savedData.groups[g]
    elseif g == -1 then
      top.data = ignoredDummy
    end
    top:Show()
    h = h+5+top:GetHeight()
    i = i+1
    if g == 0 or (not top.data.collapsed) then
      for _,t in ipairs(datas) do
        local entries = {}
        for entry in pairs(t) do table.insert(entries,entry) end
        table.sort(entries)
        for _,entry in ipairs(entries) do
          local cData = t[entry]
          local header
          local n = 0
          for _,c in ipairs(cData) do
            local group = savedData.lookup[c.spawnType][c.spawnId] or 0
            if group == g then
              if not header then
                header = children[i]
                header.groupHeader = top
                header:SetState("ENTRYHEADER")
                header.dropdown.text:SetText(top.dropdown.text:GetText())
                header.dropdown:Show()
                header.data = cData
                header:Show()
                h = h+5+header:GetHeight()
                i = i+1
              end
              if not cData.collapsed[g] then
                local entry = children[i]
                entry.groupHeader = top
                entry.entryHeader = header
                entry:SetState("SPAWNID")
                entry.text:SetText(("%s @ %d"):format(c.name, c.spawnId))
                entry.dropdown.text:SetText(top.dropdown.text:GetText())
                entry.dropdown:Show()
                entry.data = c
                entry:Show()
                h = h+5+entry:GetHeight()
                i = i+1
              end
              n = n+1
            end
          end
          if header then
            header.text:SetText(("#%05d: %s (x%d)"):format(cData[1].entry, cData[1].name, n))
          end
        end
      end
    end
  end
  while i <= #children do
    children[i]:Hide()
    i=i+1
  end
  scrollChild:SetHeight(h)
  scrollChild:SetWidth(scrollParent:GetWidth())
  scrollParent:UpdateScrollChildRect()
  scrollParent:SetVerticalScroll(math.min(scrollParent:GetVerticalScrollRange(),scrollParent:GetVerticalScroll()))
end

function frame.Redraw(force) GetDataStore(redrawFrame, force) end

local sqlEditBox = CreateFrame("Frame")
sqlEditBox:SetFrameStrata("FULLSCREEN")
sqlEditBox:SetPoint("TOPLEFT")
sqlEditBox:SetPoint("BOTTOMRIGHT")
sqlEditBox:Hide()
background(sqlEditBox,0.8,0.4,0.8)
sqlEditBox.label = sqlEditBox:CreateFontString(nil,"ARTWORK")
sqlEditBox.label:SetPoint("TOPLEFT")
sqlEditBox.label:SetPoint("TOPRIGHT")
sqlEditBox.label:SetFont("FONTS\\ARIALN.TTF", 16, "OUTLINE")
sqlEditBox.label:SetTextColor(0.85,0.85,0.85)
sqlEditBox.label:SetText("SQL output is visible below - press CTRL-C (CMD-C on Mac) to copy it to your clipboard.\nPress any character key to hide this dialog.")
sqlEditBox.box = CreateFrame("EditBox", nil, sqlEditBox)
sqlEditBox.box:SetPoint("BOTTOMRIGHT", -10, 10)
sqlEditBox.box:SetPoint("TOPLEFT", 10, -40)
sqlEditBox.box:SetAutoFocus(false)
sqlEditBox.box:SetFont("FONTS\\FRIZQT__.TTF", 14)
sqlEditBox.box:SetTextColor(0,0,0)
sqlEditBox.box:SetMultiLine(true)
sqlEditBox:EnableMouse(true)
sqlEditBox:SetScript("OnMouseUp",function() sqlEditBox:Hide() end)
sqlEditBox.box:SetScript("OnChar",function() sqlEditBox:Hide() end)
generateSQL = function()
  local groups = frame.savedData.groups
  local mapping = frame.savedData.lookup
  local lines = {}
  local function push(l,...) lines[#lines+1] = (l and l:format(...) or "") end
  local function splice() local _,c=lines[#lines]:find(".*,") if not c then return end lines[#lines] = lines[#lines]:sub(1,c-1)..';'..lines[#lines]:sub(c+1) end
  push("-- SpawnGroupMapper auto-generated SQL for map %d (%s)",frame.mapid, frame.mapname)
  push()
  push("-- Constant setter section - adjust this")
  push("-- spawn_group_template (%d entries)",#groups)
  push("SET @FIRSTGROUPID := <CHANGEME>;")
  push("-- Boss state associations")
  local constants = {}
  for i=1,#groups do
    constants[i] = groups[i].name:gsub("[^a-zA-Z]",""):upper()
    push("SET @%s := <CHANGEME>; -- boss state that controls group \"%s\"", constants[i], groups[i].name)
  end
  push()
  push("-- DB modification section - you shouldn't need to touch this")
  push("-- Insert spawn group templates")
  push("DELETE FROM `spawn_group_template` WHERE `groupId` BETWEEN @FIRSTGROUPID+0 AND @FIRSTGROUPID+%d;",#groups-1)
  push("INSERT INTO `spawn_group_template` (`groupId`,`groupName`,`groupFlags`) VALUES")
  for i=1,#groups do
    push("    (@FIRSTGROUPID+%d, \"%s - %s\", 0x04)%s",i-1,frame.mapname,groups[i].name,i == #groups and ";" or ",")
  end
  push()
  push("-- Insert spawn group data")
  push("DELETE FROM `spawn_group` WHERE `groupId` BETWEEN @FIRSTGROUPID+0 AND @FIRSTGROUPID+%d;",#groups-1)
  push("INSERT INTO `spawn_group` (`groupId`,`spawnType`,`spawnId`) VALUES")
  for spawnType=0,1 do
    for spawnId, groupId in pairs(mapping[spawnType]) do
      if groupId > 0 then
        push("    (@FIRSTGROUPID+%d, %d, %d),",groupId-1,spawnType,spawnId)
      end
    end
  end
  splice()
  push()
  
  push("-- Link information for instance script")
  push("DELETE FROM `instance_spawn_groups` WHERE `instanceMapId` = %d;",frame.mapid)
  push("INSERT INTO `instance_spawn_groups` (`instanceMapId`, `bossStateId`, `bossStates`, `spawnGroupId`, `flags`) VALUES");
  for i=1,#groups do
    push("    (%d,@%s,0x17,@FIRSTGROUPID+%d,0x1), -- Spawn %s if state is not DONE",frame.mapid,constants[i],i-1,groups[i].name)
  end
  splice()
  push()
  
  push("-- Remove legacy linked_respawn entries")
  push("DELETE lr FROM `linked_respawn` lr INNER JOIN `spawn_group` sg ON sg.`groupId` BETWEEN @FIRSTGROUPID+0 AND @FIRSTGROUPID+%d AND sg.`spawnType` = (lr.linkType >> 1) AND sg.`spawnId`=lr.`guid`;",#groups-1)
  push()
  
  sqlEditBox:Show()
  sqlEditBox.box:SetText(table.concat(lines,"\n"))
  sqlEditBox.box:SetFocus()
  sqlEditBox.box:HighlightText()
end

SlashCmdList.SPAWNGROUPMAPPER = function()
  if frame:IsShown() then
    frame:Hide()
  else
    frame:Show()
    frame:Redraw()
  end
end
SLASH_SPAWNGROUPMAPPER1 = "/spawngroup"